/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * This code is confidential and provided to Android partners under NDA. You may modify, compile and
 * distribute the compiled code with your management software to assist with the provisioning and
 * management of Android devices. Other uses require written permission.
 */
package com.google.android.apps.work.dpcsupport;

import android.content.Context;
import android.os.Handler;
import android.os.SystemClock;
import android.support.annotation.VisibleForTesting;
import android.support.annotation.WorkerThread;
import android.util.Log;
import com.google.android.apps.work.dpcsupport.WorkingEnvironmentCallback.Error;
import java.util.concurrent.TimeUnit;

/**
 * Updates Play Services using {@link PlayServicesUpdater} and retries on failures.
 */
class PlayServicesUpdateScheduler {
  /** The download size is ~50mb. */
  @VisibleForTesting
  static final long PLAY_SERVICES_UPDATE_TIMEOUT_MS =
      TimeUnit.MILLISECONDS.convert(20, TimeUnit.HOURS);

  /** Delay after which the update is not retried */
  @VisibleForTesting
  static final long PLAY_SERVICES_UPDATE_START_TIMEOUT_MS =
      TimeUnit.MILLISECONDS.convert(2, TimeUnit.MINUTES);

  @VisibleForTesting
  static final long PLAY_SERVICES_UPDATE_RETRY_WAIT_MS =
      TimeUnit.MILLISECONDS.convert(10, TimeUnit.SECONDS);

  private final Context context;
  private final Handler backgroundHandler;

  private WorkingEnvironmentCallback callback;
  private long maxPlayServicesUpdateStartTimeMs;
  private boolean updateComplete = false;

  PlayServicesUpdateScheduler(Context context, Handler backgroundHandler) {
    this.context = context;
    this.backgroundHandler = backgroundHandler;
  }

  /**
   * Updates play services to a version supporting Android for Work Accounts, and retries on
   * failures
   */
  @WorkerThread
  void updatePlayServices(WorkingEnvironmentCallback callback) {
    this.callback = callback;

    activateTimeout();

    tryUpdatePlayServices();
  }

  @WorkerThread
  private void tryUpdatePlayServices() {
    final WorkingEnvironmentCallback updateCompleteCallback =
        new WorkingEnvironmentCallback() {
          @Override
          public void onSuccess() {
            success();
          }

          @Override
          public void onFailure(Error statusCode) {
            retry(statusCode);
          }

          @Override
          public void onProgressChange(float progress) {
            callback.onProgressChange(progress);
          }
        };

    new PlayServicesUpdater(context, backgroundHandler).updatePlayServices(updateCompleteCallback);
  }

  private void activateTimeout() {
    maxPlayServicesUpdateStartTimeMs =
        SystemClock.uptimeMillis() + PLAY_SERVICES_UPDATE_START_TIMEOUT_MS;

    backgroundHandler.postDelayed(
        new Runnable() {
          @Override
          public void run() {
            failure(Error.PLAY_SERVICES_UPDATE_TIMEOUT);
          }
        },
        PLAY_SERVICES_UPDATE_TIMEOUT_MS);
  }

  private void success() {
    if (updateComplete) {
      return;
    }
    Log.i(Constants.TAG, "Play services successfully updated.");
    updateComplete = true;
    callback.onSuccess();
  }

  private void retry(Error statusCode) {
    if (updateComplete) {
      return;
    }
    if (SystemClock.uptimeMillis() > maxPlayServicesUpdateStartTimeMs) {
      failure(statusCode);
      return;
    }

    Log.i(Constants.TAG, "Retrying Play Services update.");
    backgroundHandler.postDelayed(
        new Runnable() {
          @Override
          public void run() {
            tryUpdatePlayServices();
          }
        },
        PLAY_SERVICES_UPDATE_RETRY_WAIT_MS);
  }

  private void failure(Error reason) {
    if (updateComplete) {
      return;
    }
    updateComplete = true;
    callback.onFailure(reason);
  }
}
