/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * This code is confidential and provided to Android partners under NDA. You may modify, compile and
 * distribute the compiled code with your management software to assist with the provisioning and
 * management of Android devices. Other uses require written permission.
 */
package com.google.android.apps.work.dpcsupport;

import android.app.PendingIntent;
import android.app.admin.DevicePolicyManager;
import android.content.BroadcastReceiver;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.IntentSender;
import android.content.pm.PackageInstaller;
import android.os.Bundle;
import android.os.Handler;
import android.os.UserManager;
import android.support.annotation.WorkerThread;
import android.util.Log;
import com.google.android.apps.work.dpcsupport.WorkingEnvironmentCallback.Error;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * Installs the downloaded Play Store APK.
 */
class PlayStoreInstaller extends BroadcastReceiver {

  private static final String PLAY_STORE_UPDATE_COMPLETE_INTENT = "PLAY_STORE_UPDATE_COMPLETE";

  private final Context context;
  private final ComponentName admin;
  private final Handler backgroundHandler;
  private final DevicePolicyManager devicePolicyManager;
  private final PackageInstaller packageInstaller;
  private final UserManager userManager;

  private WorkingEnvironmentCallback callback;
  private boolean previousEnsureVerifyAppsSetting;

  PlayStoreInstaller(Context context, ComponentName admin, Handler backgroundHandler) {
    this.context = context;
    this.admin = admin;
    this.backgroundHandler = backgroundHandler;
    this.devicePolicyManager =
        (DevicePolicyManager) context.getSystemService(Context.DEVICE_POLICY_SERVICE);
    this.packageInstaller = context.getPackageManager().getPackageInstaller();
    this.userManager = (UserManager) context.getSystemService(Context.USER_SERVICE);
  }

  /**
   * Installs the provided Play Store APK using Device Owner privileges, and reports the result to
   * the {@link WorkingEnvironmentCallback}.
   */
  void installPlayStore(InputStream inputStream, WorkingEnvironmentCallback callback) {
    this.callback = callback;

    context.registerReceiver(this, new IntentFilter(PLAY_STORE_UPDATE_COMPLETE_INTENT));
    hideVerifyConsentPopup();
    try {
      installPackage(inputStream);
    } catch (IOException e) {
      Log.e(Constants.TAG, "Failed to install play store apk", e);
      fail(Error.PLAY_STORE_INSTALL_FAILED);
    }
  }

  /**
   * Installs a package.
   *
   * @throws IOException when installation fails
   */
  @WorkerThread
  private void installPackage(InputStream inputStream) throws IOException {
    PackageInstaller.SessionParams params =
        new PackageInstaller.SessionParams(PackageInstaller.SessionParams.MODE_FULL_INSTALL);
    params.setAppPackageName(Constants.PLAY_STORE_PACKAGE_NAME);
    int sessionId = packageInstaller.createSession(params);
    PackageInstaller.Session session = packageInstaller.openSession(sessionId);
    OutputStream out = session.openWrite(Constants.TAG, 0, -1);
    byte[] buffer = new byte[65536];
    int numByesRead;
    while ((numByesRead = inputStream.read(buffer)) != -1) {
      out.write(buffer, 0, numByesRead);
    }
    session.fsync(out);
    inputStream.close();
    out.close();
    session.commit(createIntentSender(context, sessionId));
  }

  /**
   * Creates an {@link IntentSender} for reporting the install result.
   *
   * @param context application context
   * @param sessionId sessionId of the {@link PackageInstaller}
   * @return an {@link IntentSender} for reporting the install result
   */
  private IntentSender createIntentSender(Context context, int sessionId) {
    PendingIntent pendingIntent =
        PendingIntent.getBroadcast(
            context, sessionId, new Intent(PLAY_STORE_UPDATE_COMPLETE_INTENT), 0);
    return pendingIntent.getIntentSender();
  }

  /**
   * The first time a package is installed from an external source, a "verify consent" dialogue is
   * shown. This dialogue disturbs the setup flow. We temporarily accept app verification so that
   * this is not shown, as we know that we are installing the Play Store with known signed keys.
   */
  private void hideVerifyConsentPopup() {
    Bundle userRestrictions = userManager.getUserRestrictions();
    previousEnsureVerifyAppsSetting = userRestrictions.getBoolean(UserManager.ENSURE_VERIFY_APPS);
    if (!previousEnsureVerifyAppsSetting) {
      devicePolicyManager.addUserRestriction(admin, UserManager.ENSURE_VERIFY_APPS);
    }
  }

  private void resetVerificationPopupSettings() {
    if (!previousEnsureVerifyAppsSetting) {
      devicePolicyManager.clearUserRestriction(admin, UserManager.ENSURE_VERIFY_APPS);
    }
  }

  private void fail(Error reason) {
    context.unregisterReceiver(this);
    resetVerificationPopupSettings();
    callback.onFailure(reason);
  }

  private void success() {
    Log.i(Constants.TAG, "Successfully updated Play Store.");
    context.unregisterReceiver(this);
    resetVerificationPopupSettings();
    callback.onSuccess();
  }

  @Override
  public void onReceive(Context context, Intent intent) {
    // Run on background thread.
    backgroundHandler.post(
        new Runnable() {
          @Override
          public void run() {
            success();
          }
        });
  }
}
